// Hyperion C-to-Z80 Compiler by Mikolaj Franaszczuk. 
// Copyright 1999. Version 1.032 August 31, 1999
// E-mail msf22@cornell.edu with questions or comments
// Visit the Hyperion web page at: http://www.geocities.com/TimesSquare/Arcade/2505/hyperion.html
// This source file may not be modified and\or distributed without prior permission
// A compilation of this or a modified source file may not be distributed without prior permission

// Find main() at the bottom of this file for main program structure

#include <string.h>
#include <fstream.h>
#include <stdlib.h>
#include <conio.h>
#include <stdio.h>
#include <ctype.h>

#define MAX_LEN 100
// Maximum length of any command line

#define M_FLIP {infile.close();outfile.close();work3=work;work=work2;work2=work3;infile.open(work);outfile.open(work2);}
// M_FLIP flips the work files so that infile is always new input
//   and outfile is always new output. Opens the files appropriatly

#define M_GL infile.getline(cl.txt,MAX_LEN);
// M_GL Reads line from infile to cl

#define M_PLT PutLine(outfile,temp)
// M_PLT Puts line temp into outfile

#define M_PLC PutLine(outfile,cl)
// M_PLC Puts line cl into outfile

#define M_CF cout.flush()
// M_CF Empties the standard output buffer (displays stuff on screen)

class ZString
{
// This is the super-string class object for the command lines
public:
	ZString* oplist;
	char* txt;
	// txt holds the actual characters of the string
	ZString();
	ZString(char*);
	ZString(int);
	operator =(char*);	
	operator =(ZString);
	operator =(int);
	ZString operator +(char*);
	ZString operator +(ZString);
	ZString operator +(int);
	bool operator ==(char*);
	bool operator ==(ZString);
	bool operator ==(int);
	bool operator !=(char*);
	bool operator !=(ZString);
	bool operator !=(int);
	int Len();
	int Find(char*,int,int);
	int Find(char*);
	int Find(ZString,int,int);
	int Find(ZString);
	ZString Cut(int,int);
	Insert(ZString,int);
	Replace(ZString,ZString);
	DelSpaces();
	DelAllChars(char);
	Capitalize();
	UnTab();
	ZString MidFT(int,int);
	ZString Mid(int,int);
	bool Cmnd(char*);
	int IsEqual();
	int IsFunc();
	int IsPar();
	int IsAri();
	int Flag();
	int IsOper(int);
	int FindMatch(int);
	int SLeftForOp(int);
	int SRightForOp(int);
	Disp();
	Disp(int);
};

void EndProgram(); // Routine that runs when an error is encountered
void MOverflow(); // Executes when compiler runs out of memory
int IntegrityCheck(ZString); // Checks number of parentheses in command lines
int GetRT(ZString); // Gets return types of functions
ZString ZSCon(int); // These two functions convert 
ZString ZSCon(char*); // anything to a ZString
ZString Ops[14]; // Array of operators, parentheses etc.
ZString parname[10]; // Parameter name table
ZString cl; // Main command line
ZString cline; // Error-marker original source code line
ZString temp,p1,p2,tempar[10]; // Supplementary ZStrings
ifstream infile; // Input file
ofstream outfile; // Output file
ZString SFile; // Physical source file
ZString DFile; //Physical destination file
ZString fname[10]; // Names of included files
ZString var[400]; // Variable symbolic table
ZString ustr[50]; // String symbolic table
ZString fntable[100]; // Function name table
ZString reg[15]; // Names of registers
ZString arname[50]; // Names of arrays and lists
ZString nestlabel[20]; // For saving nest related label info
ZString nestgoto[20]; // For saving nest related label info
ZString title; // Program title
bool debug; // Indicator of whether compilation is in debug mode
char* work="work.dat"; // Starting input file
char* work2="work2.dat"; // Starting output file
char* work3="untitled.txt"; // Temporary file for M_FLIP macro
int aorl[50]; // Indicator of whether an arname is a list or array
int vtotal; // Total number of variables
int vtype[400]; // Variable type array
int vstat[400]; // Indicates whether variable is permanent or temporary
int a,b,c; // Temporary calculation variables
int nw; // Indicator at multiple spots of whether a line should be
	    //   copied directly or not
int mp[500]; // Temporary internal memory pointer
int cn=1; // Current nest level indicator
int nesttype[20]; // Type of nest
int lc; // Line counter for pre-parser and mem-block counter for parser
int nstr=0; // String counter
int nars=0; // Number of arrays and lists
int arx[50]; // Row sizes of arrays
int ary[50]; // Column sizes of arrays
int nf; // Number of declared functions counter 
int fttable[100]; // Function type table
int mt[500]; // Temporary memory type (1 or 2 byte)
int lcoms; // Toggle of whether original lines should be kept as comments
int ip; // Numnber of function parameters
int partype[10]; // Parameter type (1 or 2 byte)
int fptable[100][10]; // Parameter types for all functions
int inf; // Toggle of whether parser is in function declaration
int ipars[100]; // Number of parameters for each function
int crt; // Current function return type
int pc; // Parenthese counter for integrity checker
int parsing=0; // Indicator of whether parsing has begun
int preparsing=0; // Indicator of whether pre-parsing has begun
int fcalled[100]; // Array of whether a function was called or not
int retexist; // Flag of whether function returned a value
int tl; // Total label counter
int tinc=0; // Number of include files
int lcn; // Nest number of latest FOR or LOOP
int ustrlen[50]; // Stores lengths of ustr strings
int ilv=0; // Number of label variables
int lv[50]; // Variable numbers of label variables
int chksum=0; // Indicator of whether permanent memory altering instructions were used

ZString::ZString()
{
	// Construct an empty ZString
	txt = new char[MAX_LEN];
	if(txt==0)MOverflow();
	txt[0] = 0;
	oplist=Ops;
}
 
ZString::ZString(char* s)
{
	// Construct a ZString based on a string
	txt = new char[MAX_LEN];
	if(txt==0)MOverflow();
	strcpy(txt,s);
	oplist=Ops;
}

ZString::ZString(int i)
{
	// Construct a ZString based on a number
	char s[5];
	txt = new char[MAX_LEN];
	if(txt==0)MOverflow();
	itoa(i,s,10);
	strcpy(txt,s);
	oplist=Ops;
}

ZString::operator=(char* s)
{
	// Make a ZString equal a string
	strcpy(txt,s);
}

ZString::operator=(ZString s)
{
	// Make the ZString equal another one
	strcpy(txt,s.txt);
}

ZString::operator=(int i)
{
	// make a ZString equal a number
	char s[MAX_LEN];
	itoa(i,s,10);
	strcpy(txt,s);
}

ZString ZString::operator+(char* s)
{
	// Appends a string to a ZString
	ZString ret;
	strcat(ret.txt,txt);
	strcat(ret.txt,s);
	return ret;
}

ZString ZString::operator+(ZString s)
{
	// Appends one ZStrting to another
	ZString ret;
	strcpy(ret.txt,txt);
	strcat(ret.txt,s.txt);
	return ret;
}

ZString ZString::operator+(int i)
{
	// Appends a number to a ZString
	char s[5];
	ZString ret;
	itoa(i,s,10);
	strcpy(ret.txt,txt);
	strcat(ret.txt,s);
	return ret;
}

bool ZString::operator==(char* s)
{
	// Compares a ZString to a string
	if(Len()<0)return false;
	int r1,r2;
	r1 = strncmp(txt,s,Len()+1);
	r2 = strncmp(txt,s,strlen(s));
	if((r1==0)&&(r2==0))return true;
	else return false;
}

bool ZString::operator!=(char* s)
{
	if((*this)==s)return false;
	else return true;
}

bool ZString::operator==(ZString s)
{
	// Compares one ZString to another
	if(Len()<0)return false;
	int r1,r2;
	r1 = strncmp(txt,s.txt,Len()+1);
	r2 = strncmp(txt,s.txt,s.Len()+1);
	if((r1==0)&&(r2==0))return true;
	else return false;
}

bool ZString::operator!=(ZString s)
{
	if((*this)==s)return false;
	else return true;
}

bool ZString::operator==(int i)
{
	// Compares a ZString to an integer
	if(Len()<0)return false;
	int r1, r2;
	char s[5];
	itoa(i,s,10);
	r1 = strncmp(txt,s,Len()+1);
	r2 = strncmp(txt,s,strlen(s));
	if((r1==0)&&(r2==0))return true;
	else return false;
}

bool ZString::operator!=(int i)
{
	if((*this)==i)return true;
	else return false;
}

int ZString::Len()
{
	// Returns one less than length of string
	// It is the last character offset of the ZString
	return strlen(txt)-1;
}

bool ZString::Cmnd(char* s)
{
	// Checks if first characters of the ZString are s
	if((Len()>=(signed)strlen(s)-1)&&(MidFT(0,strlen(s)-1)==s))return true;
	return false;
}

int ZString::Find(ZString s)
{
	return Find(s.txt,0,1);
}

int ZString::Find(ZString s,int a, int b)
{
	return Find(s.txt,a,b);
}

int ZString::Find(char* token)
{
	return Find(token,0,1);
}

int ZString::Find(char* token, int start, int dir)
{
	// Find a given string in a ZString starting at start
	ZString t(token);
	int a;
	if(start<=Len())
	{
		if(dir==1)
		{
			// Searching right
			for(a=start;a<=(Len()-t.Len());a++)
			{
					int n=0;
					for(int b=a;b<=a+t.Len();b++)
						if(t.txt[b-a]!=txt[b])n=1;
					if(n==0)return a;
			}
			return -1;
		}
		else
		{
			// Searching left
			for(a=start;a>=0;a--)
			{
				int n=0;
				for(int b=a-t.Len();b<=a;b++)
					if(t.txt[b-a+t.Len()]!=txt[b])n=1;
				if(n==0)return a-t.Len();
			}
			return -1;
		}
	}
	return -1;
	// Returns location or -1 of not found
}

ZString ZString::MidFT(int from, int to)
{
	// Gets the middle part of a ZString from a spot to a spot
	ZString ret;
	int k=0;
	for(int a=from;a<=to;a++)
	{
		ret.txt[k]=txt[a];
		k++;
	}
	ret.txt[k]=0;
	return ret;
}

ZString ZString::Mid(int start, int length)
{
	// Gets the middle part of a string from start to a given length
	ZString ret;
	int k=0;
	for(int a=start;a<start+length;a++)
	{
		ret.txt[k]=txt[a];
		k++;
	}
	ret.txt[k]=0;
	return ret;
}

ZString ZString::Cut(int from, int to)  
{
	// Cuts part of a string, returning it, and leaves the remainder
	ZString ret;
	ret = MidFT(from,to);
	(*this) = MidFT(0,from-1) + MidFT(to+1,Len());
	return ret;
}

ZString::Disp()
{
	// Displays the ZString without leading or trailing new lines
	cout << txt;M_CF;
}

ZString::Disp(int o)
{
	//Displays the ZString
	if(o==1)
	{
		// With leading new line
		cout << '\n' << txt;M_CF;
	}
	if(o==2)
	{
		// With trailing new line
		cout << txt << '\n';M_CF;
	}
	if(o==3)
	{
		// With leading and trailing new lines
		cout << '\n' << txt << '\n';M_CF;
	}
}

int ZString::IsEqual()
{
	// Checks if the ZString has an equal sign in it
	int b = 0;
	int a;
again:
	a = Find("=",b,1);
	if(a!=-1)
	{
		// Must verify that the supposed equal sign isn't really ==
		b = Find("==",b,1);
		if(a!=b)return 1;
		b=b+2;
		if(b<=Len())goto again;
	}
	return 0;
}

int ZString::IsFunc()
{
	// Checks if the ZString has a function in it
	int a=0;
again:
	a = Find("(",a,1);
	if(a!=-1)
	{
		// Must verify if it is really a function, not just
		//   a set of parentheses. This is accomplished by checking
		//   if there is an operator to the left (there shouldn't be one)
		if(IsOper(a-1)==0)return 1;
		a++;
		if(a<=Len())goto again;
	}
	return 0;
}

int ZString::IsPar()
{
	// Checks if the ZString has a set of parentheses in it
	int a=0;
again:
	a = Find("(",a,1);
	if(a!=-1)
	{
		// Must verify that they are really parentheses, not
		//   a function. This is accomplisehd by checking if there
		//   is an operator the the left (there should be one)
		if(IsOper(a-1)==1)return 1;
		a++;
		if(a<=Len())goto again;
	}
	return 0;
}

int ZString::IsAri()
{
	// Checks if there are arithmetic of logical operators
	//   in the ZString
	for(int a=0;a<10;a++)
		if(Find(oplist[a])!=-1)return 1;
	return 0;
}

int ZString::Flag()
{
	// Returns a unique flag value based on the type of stuff in
	//  the ZString. 
	int ret=0;
	if(IsEqual()==1)ret++;
	if(IsFunc()==1)ret=ret+2;
	if(IsPar()==1)ret=ret+4;
	if(IsAri()==1)ret=ret+8;
	return ret;
}

int ZString::IsOper(int x)
{
	// Checks if a given character is an operator ( or comma etc.)
	for(int a=0;a<14;a++)
		if(txt[x]==oplist[a].txt[0])return 1;
	return 0;
}

int ZString::FindMatch(int s)
{
	// Find a the closing parentheses for an opening one at the
	//  given location
	int c=0;
	for(int a=s;a<=Len();a++)
	{
		if(txt[a]=='(')c++;
		if(txt[a]==')')c--;
		// Takes care of other nested parentheses
		if(c==0)break;
	}
	return a;
}

ZString::DelAllChars(char c)
{
	// Removes all given characters from ZString
	ZString temp;
	int k=0;
	for(int a=0;a<=Len();a++)
		if(txt[a]!=c)
		{
			temp.txt[k]=txt[a];
			k++;
		}
	temp.txt[k]=0;
	(*this)=temp;
}

ZString::DelSpaces()
{
	// Removes all spaces from ZString
	DelAllChars(' ');	
}

ZString::Replace(ZString o, ZString n)
{
	// Replaces a set of characters (o) with a different set (n)
	//   in the ZString
	int a;
again:
	a = Find(o);
	if(a!=-1)
	{
		// If o was found, replace it with n
		Cut(a,a+o.Len());
		(*this)=Mid(0,a)+n+MidFT(a,Len());
		goto again;
		// Keep searching the same ZString for more "o"s...
	}
}

ZString::Insert(ZString s, int w)
{
	// Inserts a ZString into the current ZString at a given 
	//   location
	if(w>0)
		(*this)=MidFT(0,w-1)+s+MidFT(w,Len());
	else
		(*this)=s+(*this);
}

ZString::Capitalize()
{
	// Capitalizes all the letters in the ZString
	strupr(txt);
}

ZString::UnTab()
{
	// Removes all leading spaces from the ZString
	for(int a=0;a<=Len();a++)
	{
		if((txt[a]!=' ')&&(txt[a]!=9))break;
	}
	(*this)=MidFT(a,Len());
}

int ZString::SLeftForOp(int s)
{
	c=0;
	for(a=s-1;a>=0;a--)
	{
		// Scan each character left of start position...
		for(b=0;b<13;b++)
			if(txt[a]==oplist[b].txt[0])c=1;
			// Breaks if the character is an operator or comma
		if(c==1)break;	
	}
	if(c==0)return -1;
	return a;
}

int ZString::SRightForOp(int s)
{
	c=0;
	for(a=s+1;a<=Len();a++)
	{
		// Scan each character right of start position...
		for(b=0;b<14;b++)
			if(txt[a]==oplist[b].txt[0])c=1;
			// Breaks if the character is an operator or comma
		if(c==1)break;	
	}
	if(c==0)return Len()+1;
	return a;	
}

ZString ZSCon(int i)
{
	// Converts a number into a ZString
	ZString ret;
	itoa(i,ret.txt,10);
	return ret;
}

ZString ZSCon(char* s)
{
	// Converts a string into a ZString
	ZString ret;
	ret.txt = s;
	return ret;
}



void PutLine(ofstream f, ZString s)
{
	// Puts a ZString into a file as a line
	f.write(s.txt,s.Len()+1);
	f.put('\n');
}

void PutLine(ofstream f, char* s)
{
	// Puts a string into a file as a line
	f.write(s,strlen(s));
	f.put('\n');
}

void SetOps()
{
	// Initialization procedures
	// Sets the operator array 
	Ops[0] = "==";
	Ops[1] = "<>";
	Ops[2] = "<";
	Ops[3] = ">";
	Ops[4] = "*";
	Ops[5] = "/";
	Ops[6] = "+";
	Ops[7] = "-";
	Ops[8] = "&&";
	Ops[9] = "||";
	Ops[10] = "(";
	Ops[11] = "=";
	Ops[12] = ",";
	Ops[13] = ")";
	// Defines Hyperion functions 
	nf=14;
	fntable[0] = "IF";
	fntable[1] = "FOR";
	fntable[2] = "GOTO";
	fntable[3] = "WHILE";
	fntable[4] = "RETURN";
	fntable[5] = "PRINT";
	fntable[6] = "SPRITE";
	fntable[7] = "LOAD";
	fntable[8] = "SAVE";
	fntable[9] = "ARGET";
	fntable[10] = "ARPUT";
	fntable[11] = "BYTEARRAY";
	fntable[12] = "LIST";
	fntable[13] = "LGET";
	fntable[14] = "LPUT";
	// Defines Hyperion function return types
	fttable[0]=0;
	fttable[1]=0;
	fttable[2]=0;
	fttable[3]=0;
	fttable[4]=0;
	fttable[5]=0;
	fttable[6]=0;
	fttable[7]=2;
	fttable[8]=0;
	fttable[9]=1;
	fttable[10]=0;
	fttable[11]=0;
	fttable[12]=0;
	fttable[13]=1;
	fttable[14]=0;
	// Defines number of parameters for each Hyperion function
	ipars[0]=1;
	ipars[1]=4;
	ipars[2]=1;
	ipars[3]=1;
	ipars[4]=1;
	ipars[5]=3;
	ipars[6]=1; 
	ipars[8]=2;
	ipars[9]=3;
	ipars[10]=4;
	ipars[11]=3;
	ipars[12]=1;
	ipars[13]=2;
	ipars[14]=3;
	// Defines register names
	reg[5]="B";
	reg[6]="C";
	reg[7]="D";
	reg[8]="E";
	reg[9]="H";
	reg[10]="L";
	reg[11]="BC";
	reg[12]="DE";
	reg[13]="HL";
	for(a=1;a<20;a++)
	{
		// "Zeroes" initial nest types
		nesttype[a]=3;
		// Sets all temp mem variable types to double
		mt[a]=2;
	}
	title = "No description given.";
}

void AssignVar(ZString cl,int v)
{
	// Registers variables in the symbolic table
	if(cl.Find("(")==-1)
	{
		// Only assign it if it's really a variable, not
		//   a function declaration
		if(vtotal>375)
		{
			cout << "\nMaximum number of variables exceeded.";
			EndProgram();
		}
		a=0;
		b=0;
		int aa;
		while(b==0)
		{
			a = cl.Find(",",6,1);
			aa=0;
			// Find commas that separate variable names
			if(a==-1)
			{
				// If no more are found, mark it as the last
				//   variable
				b=1;
				a=cl.Len();
			}
			temp = cl.Cut(6,a);
			if(b==0)
				temp = temp.MidFT(0,temp.Len()-1);
			if(!isalpha(temp.txt[0]))
			{
				cout << "\nA variable name must begin with a letter.";
				EndProgram();
			}
			if(temp.txt[temp.Len()]=='$')
			{
				// A $ at the end indicates a permanent variable
				aa=1;
				chksum=1;
				temp = temp.MidFT(0,temp.Len()-1);	
			}
			if(temp.Cmnd("PARV")||temp.Cmnd("MEM"))
			{
				cout << "\nIllegal variable name (PARV, MEM, and RETV are not allowed).";
				EndProgram();
			}
			if(temp.Len()<1)
			{
				cout << "\nA variable name must be at least 2 letters long.";
				EndProgram();
			}
			if(temp.Len()>20)
			{
				cout << "\nA variable name must contain less than 20 letters.";
				EndProgram();
			}
			// First check if a variable by this name doesn't
			//   already exist
			for(c=1;c<=vtotal;c++)
				if(var[c]==temp)
				{
					cout << "\nThe variable ";
					M_CF;
					var[c].Disp(2);
					cout << " has already been declared.";
					EndProgram();
				}
			vtotal++;
			var[vtotal]=temp;
			// Sets new variable name
			vtype[vtotal]=v;
			vstat[vtotal]=aa;
			// Variable type is its return size (1 or 2)
		}
		nw=1;
	}
}

int GetRT(ZString s)
{
	// Find and returns the given function string's return type
	for(int a=0;a<=nf;a++)
	{
		if(fntable[a]==s)
		{
			return fttable[a];
		}
	}
	cout << "\nFunction does not exist!\n";
	M_CF;
	s.Disp(2);
	EndProgram();
	return 0;
}

void PrepareFile(ZString SFile)
{
	// Prepares source file for the pre-parser
	infile.open(SFile.txt,ios::nocreate);
	if(!infile.is_open())
	{
		cout << "\nSource file not found.";
		M_CF;
		EndProgram();
	}
	DFile = SFile.MidFT(0,SFile.Len()-3)+"asm";
	// Sets name of destination file
	outfile.open(work2);
	temp = "SINGLE RETV1";M_PLT;
	temp = "DOUBLE RETV2, OPVAR1, OPVAR2, TEMPM";M_PLT;
	// Inserted internal Hyperion variable declarations at top of file
	while(infile)
	{
		// Copies source file to work file, removing leading spaces
		//   and removing # symbols
		M_GL;
		cl.UnTab();
		temp=cl;
		temp.DelSpaces();
		if(!isgraph(temp.txt[0])&&temp.Len()>0)
		{
			// Some weird non-alphanumeric character was found
			preparsing=0;
			cout << "\nNon-compilable file. Invalid file format.";
			EndProgram();
		}
		if(cl.txt[0]=='#')cl=cl.MidFT(1,cl.Len());
		if((cl.txt[0]!='\'')&&(cl.txt[0]!='@'))cl.DelAllChars(';');
		// Removes all semicolons
		M_PLC;
	}
}

void PreCompile(ifstream infile, ofstream outfile)
{
moreinc:
	lc=0;
	M_FLIP;
	// This first loop will add all the include files to the main one
	while(infile)
	{
		lc++;
		M_GL;
		p1 = cl;
		cl.DelSpaces();
		cl.Capitalize();
		// Original line was saved in p1, while cl was prepped for "INCLUDE" read
		if(cl.Cmnd("INCLUDE"))
		{
			// Load another file's code into this location
			temp = cl.MidFT(7,cl.Len());
			for(a=1;a<=tinc;a++)
				// Checks if this file already has been included
				if(fname[a]==temp)
					goto skipit;		
			tinc++;
			fname[tinc]=temp;
			infile.close();
			infile.open(temp.txt,ios::nocreate);
			if(!infile.is_open())
			{
				cout << "\nInclude file not found.\n";
				M_CF;
				temp.Disp(2);
				EndProgram();
			}
			while(infile)
			{
				// Output the loaded file line by line to outfile
				M_GL;
				// File prep work done on loaded file...
				cl.UnTab();
				if(cl.txt[0]=='#')cl=cl.MidFT(1,cl.Len());
				M_PLC;
			}
			infile.close();
			infile.open(work);
			// Locate the old position in the infile
			for(int q=1;q<=lc;q++)
				M_GL;
			while(infile)
			{
				// Place the rest of the infile behind the loaded file
				M_GL;
				M_PLC;
			}
			// The infile must be re-read
			goto moreinc;
		}
		else
		{
			// If not including a file, just copy the line directly
			cl=p1;
			M_PLC;
		}
skipit:;
	}
doover:
	cout << "*";
	M_CF;
	M_FLIP;
	lc=0;
	while(infile)
	{
		lc++;
		M_GL;
		if(_kbhit())
		{
			// User keystroke processing
			cout << "\nPre-Parsing paused. Press escape to quit, any other key to continue.";
			M_CF;
			getch();
			a=getch();
			if(a==27)
			{
				preparsing=0;
				cout << "\nPre-Parsing terminated by user.";
				EndProgram();
			}
			cout << "\nPre-Parsing resumed.\nPre-Parsing ";
			M_CF;
		}
		nw=0;
		if((cl.txt[0]!='\'')&&(cl.txt[0]!='@')&&(cl.txt[cl.Len()]!=':'))
		// If it's an assembler line or comment or label, leave as is
		{
		if((cl.txt[cl.Len()]=='{')&&(cl.Len()>0))
		{
			// Place ending {s in new lines
			PutLine(outfile,cl.MidFT(0,cl.Len()-1));
			PutLine(outfile,"{");
			nw=1;
		}
		else
		{
		// Check to see if there are quotation marks
		a=cl.Find("\"");
		if(a>-1)
		{
moremarks:
			//There are. Find the ending quotation mark
			b=cl.Find("\"",a+1,1);
			if(b==-1)
			{
				//Missing a quotation mark
				cout << "\nMissing quotation mark.";
				EndProgram();
			}
			// Cut the quotation phrase out
			temp = cl.Cut(a,b);
			int aa=a;
			temp = temp.MidFT(1,temp.Len()-1);
			nstr++;
			ustr[nstr]=temp;
			ustrlen[nstr]=b-a-1;
			if(ustrlen[nstr]>30)
			{
				temp.Disp(1);
				cout << "\nMaximum string length exceeded.";
				EndProgram();
			}
			temp = ZSCon("DOUBLEUSTR") + nstr;
			AssignVar(temp,2);
			vtype[vtotal]=4;
			// Make it seem like USTR is not a variable but a 2 byte number
			a=aa;
			// Made USTR a fake variable
			temp = ZSCon("USTR")+nstr;
			cl.Insert(temp,a);
			a=cl.Find("\"");
			if(a!=-1)goto moremarks;
			cl.DelSpaces();
			M_PLC;
			// The phrase was placed in a string array
			//   and its future address replaced it
		}
		temp=cl;
		temp.Capitalize();
		if(temp.Cmnd("TITLE"))
		{
			// Assign the Z-Shell program description
			title = cl.MidFT(5,cl.Len());
			nw=1;
		}
		if(cl.Cmnd("DEFINE"))
		{
			// Process user alias for a command
			a = cl.Find(",");
			p1 = cl.MidFT(7,a-1);
			p2 = cl.MidFT(a+1,cl.Len());
			p2.UnTab();
			p1.Capitalize();
			// Placed the alias into p1 and the substitute into p2
			while(infile)
			{
				// Go through the file and replace alias with substitute
				M_GL;
				if(cl.txt[0]!='\'')
				{
					// Don't replace in comments
					cl.Capitalize();
					if(cl.Find(p1)!=-1)
						cl.Replace(p1,p2);
				}
				M_PLC;
			}
			// Source file must be re-read
			goto doover;
		}
		cl.Capitalize();
		cl.DelSpaces();
		// The command line is prepped for the "pre-parser" at last
		if(cl.Cmnd("ASM"))
		{
			// Process assembler block
			M_GL;
			// Get ride of unnecessary {
			while(1)
			{
				// Keep reading and placing @ in front of each line
				M_GL;
				temp=cl;
				if(temp.txt[0]=='}')break;
				// End this when } is reached
				cl=ZSCon("@")+cl;
				M_PLC;
			}
			nw=1;
		}
		if(cl.Cmnd("LINECOMMENTS"))
		{
			nw=1;
			lcoms=1;
		}
		if(cl.Cmnd("SINGLE"))
			// Add single variable to symbolic table
			AssignVar(cl,1);
		if(cl.Cmnd("DOUBLE"))
			// Add double variable to symbolic table
			AssignVar(cl,2);
		}
		}
		if(nw==0)M_PLC;
		// If no overrides were written, paste the new command line
		//   into the output file
	}
}

void PushUpPars()
{
	// Pushes up parentheses in parser
	int st=0;
mpars:
	int aa=cl.Find("(",st,1);
	if((aa>0)&&(cl.IsOper(aa-1)==0))
	{
		// Ensure that these are really parentheses, not a function
		st++;
		goto mpars;
	}
	b=cl.FindMatch(aa);
	a=aa;
	temp=cl.Cut(a,b);
	temp=temp.MidFT(1,temp.Len()-1);M_PLT;
	// Pushed parentheses up after physically 
	//   removing them
	mp[lc]++;
	mt[lc]=2;
	temp = ZSCon("MEM") + ZSCon(mp[lc]) + ZSCon("=RETV2");M_PLT;
	//  Generated the memx=retv2 line and outputted it
	temp=ZSCon("MEM") + mp[lc];
	cl.Insert(temp,a);
	// Inserted the mem equivalent into the original
	//   command line
	M_PLC;
	mp[lc]++;
	nw=1;
	// Outputted revised line. Upped mem counter by 1
	//   because 2 byte memory always used. 
}

int PushUpFunc()
{
	// Pushes up function in parser
	int aa,bb,cc;
	while(1)
	{
		// Search backwards to find open parentheses
		aa = cl.Find("(",cl.Len(),-1);
		if(cl.IsOper(aa-1)==0)break;
		aa++;
		// Locates an actual function, not just parentheses
	}
	for(bb=aa-2;bb>0;bb--)
		// Locates start of function name
		if(cl.IsOper(bb)==1)break;
	bb++;
	cc=cl.FindMatch(aa);
	// Compensated location offset and found closing )
	if((cc==cl.Len())&&(bb==1));
	else
	{
		a=aa;
		b=bb;
		c=cc;
		// Pop up only if there is other stuff
		//   besides the function in the command line
		temp = cl.MidFT(b,c);M_PLT;
		// Pushed function up
		mp[lc]++;
		temp = cl.MidFT(b,aa-1);
		a=GetRT(temp);
		temp = cl.Cut(b,c);
		mt[lc]=a;
		// Increased temp mem pointer and found function
		//   return type
		temp=ZSCon("MEM") + ZSCon(mp[lc]) + ZSCon("=RETV") + a;M_PLT;
		// Generated mem=return line and outputted it
		temp=ZSCon("MEM") + mp[lc];
		cl.Insert(temp,b);
		// Inserted the mem equivalent into the orignal
		//   command line
		M_PLC;
		if(a==2)mp[lc]++;
		nw=1;
		return 1;
	}
	return 0;
}

void PushUpOps()
{
	// Pushes highest operation up for parser
	int bb;
	for(int ot=0;ot<10;ot++)
	{
		// Searches for highest operator
		bb=cl.Find(Ops[ot]);
		if(bb!=-1)break;
	}
	int tbb = bb;
	// The next two ifs ensure * or \ sign is leftmost
	if(ot==4)
	{
		bb=cl.Find(Ops[5]);
		if((bb!=-1)&&(bb<tbb))
			tbb=bb;
	}
	if(ot==5)
	{
		bb=cl.Find(Ops[4]);
		if((bb!=-1)&&(bb<tbb))
			tbb=bb;
	}
	// The next two ifs ensure + or - sign is leftmost
	if(ot==6)
	{
		bb=cl.Find(Ops[7]);
		if((bb!=-1)&&(bb<tbb))
			tbb=bb;
	}
	if(ot==7)
	{
		bb=cl.Find(Ops[6]);
		if((bb!=-1)&&(bb<tbb))
			tbb=bb;
	}
	bb=tbb;
	int aa=cl.SLeftForOp(bb);
	// The following compensates for operators with more than one
	//   character (==, <>, ||, &&)
	if((ot<2)||(ot>7))bb++; 
	int cc=cl.SRightForOp(bb);
	// Found the left and right operands
	temp=cl.Cut(aa+1,cc-1);M_PLT;
	// Cuts and pastes the operation
	mp[lc]++;
	mt[lc]=2;
	temp = ZSCon("MEM") + ZSCon(mp[lc]) + ZSCon("=RETV2");M_PLT;
	temp = ZSCon("MEM") + mp[lc];
	// Generated memx = retv2 line and puts it in outfile
	cl.Insert(temp,aa+1);
	M_PLC;
	// Inserted mem equivalent into original line and pasted it
	nw=1;
}

void ProcessKeyword()
{
	// Processes a keyword for the parser
	if(cl==ZSCon("DO"))
	{
		// Begin a do-while loop
		while(cl!="{")
			M_GL;
		cn++;
		tl++;
		nestlabel[cn] = ZSCon("NESTLBL") + tl;
		temp = nestlabel[cn] + ":";
		M_PLT;
		nesttype[cn]=0;
		return;
	}
	if(cl==ZSCon("ELSE"))
	{
		// Process "else" in if-else statement
		if(nesttype[cn]==5)
		{
			// "If" wasn't processed yet so just re-write "else"
			cn++;
			M_PLC;
		}
		else
		{
			tl++;
			p1 = ZSCon("NESTLBL") + tl;
			temp = ZSCon("GOTO(") + p1 + ")";M_PLT;
			temp = nestlabel[cn] + ":";M_PLT;
			nestlabel[cn]=p1;
		}
		return;
	}
	if(cl==ZSCon("EXIT"))
	{
		// Tranlate as assembler exit
		if(inf==1)
		{
			cout << "\nError: EXIT cannot be called from within a function.";
			EndProgram();
		}
		PutLine(outfile, "@ret");
		nw=1;
		return;
	}
	if(cl=="{")
	{
		// Indicate start of unprocessed nest
		cn++;
		nesttype[cn]=5;
		M_PLC;
		return;
	}
	if(cl=="}")
	{
		// Close of a nest
		nw=1;
		switch(nesttype[cn])
		{
		case 0:
			// End of do-while loop
			// Do nothing
			cn++;
			 break;
		case 1:
			// End of function declaration
			if(retexist==0)
			{
				cout << "\nFunction must return a value.";
				EndProgram();
			}
			temp = "@ret";M_PLT;
			temp = nestlabel[cn] + ":";M_PLT;
			inf=0;
			break;
		case 2:
			// End of FOR loop
			temp = ZSCon("@JUMP_(") + nestgoto[cn] + ")";M_PLT;
			temp = nestlabel[cn] + ":";M_PLT;
			break;
		case 3:
			// This nest indicator is a dummy, indicating the nest
			//   opener exists but has not yet been fully processed
			M_PLC;
			cn++;
			break;
		case 4:
			// End of IF nest
			temp = nestlabel[cn] + ":";M_PLT;
			break;
		case 5:
			// End of an unprocessed nest - keep } there
			M_PLC;
			break;
		default:
			cout << "\n Found a } without matching {.";
			EndProgram();
		}
		nesttype[cn]=3;
		cn--;
		return;
	}
	if(cl.Len()!=-1)
	{
		cout << "\nUnknown command.\n";
		EndProgram();
	}
}

int GetFunPars(ZString s, ZString parname[10])
{
	// Retrieves function parameters from s
	int i=0,o=0;
	int cc;
	int bb=s.Find("(");
	if(s.txt[bb+1]==')')return 0;
	while(o==0)
	{
		cc=s.Find(",",bb+1,1);
		if(cc==-1)
		{
			cc=s.Find(")",bb+1,1);
			o=1;
		}
		i++;
		if(i>9)
		{
			cout << "\nMaximum number of 9 parameters exceeded.\n";
			EndProgram();
		}
		parname[i]=s.MidFT(bb+1,cc-1);
		bb=cc;
	}
	return i;
}
int ProcessFunDecs()
{
	int dec=0;
	if(cl.Cmnd("SINGLE"))
	{
		// Single function declaration
		dec=1;
		crt=1;
		retexist=0;
	}
	if(cl.Cmnd("DOUBLE"))
	{
		// Double function declaration
		dec=1;
		crt=2;
		retexist=0;
	}
	if(cl.Cmnd("VOID"))
	{
		// Void function declaration
		dec=1;
		retexist=1;
		crt=0;
	}
	if(dec==1)
	{
		// Some type of function is beging declared
		int vt; // Valid parameter type toggle
		if(inf==1)
		{
			cout << "\nCannot declare a function from inside a function.";
			EndProgram();
		}
		inf=1;
		nw=1;
		a=cl.Find("(");
		temp=cl.MidFT(6,a-1);
		if(crt==0)temp=cl.MidFT(4,a-1);
		nf++;
		fntable[nf]=temp;
		fttable[nf]=1;
		// Assigned function to the table
		temp=ZSCon("'FSTART") + nf;M_PLT;
		temp=ZSCon("GOTO(FEND") + ZSCon(nf) + ")";M_PLT;
		temp = ZSCon("FSTART") + ZSCon(nf) + ":";M_PLT;
		// Paste appropriate labels and gotos to outfile
		cn++;
		nesttype[cn]=1;
		nestlabel[cn]=ZSCon("FEND")+nf;
		// Initiated a new nest
		ip=GetFunPars(cl,parname);
		for(b=1;b<=ip;b++)
		{
			// Gets each parameter name and type
			vt=0;
			if(parname[b].Cmnd("SINGLE"))
			{
				if(parname[b].Len()<7)
				{
					cout << "\nA parameter name must be at least two letters long.";
					EndProgram();
				}
				if(!isalpha(parname[b].txt[6]))
				{
					cout << "\nA parameter name must begin with a letter.";
					EndProgram();
				}
				parname[b]=parname[b].MidFT(6,parname[b].Len());
				partype[b]=1;
				fptable[nf][b]=1;
				vt=1;
			}
			if(parname[b].Cmnd("DOUBLE"))
			{
				if(parname[b].Len()<7)
				{
					cout << "\nA parameter name must be at least two letters long.";
					EndProgram();
				}
					if(!isalpha(parname[b].txt[6]))
				{
					cout << "\nA parameter name must begin with a letter.";
					EndProgram();
				}
				parname[b]=parname[b].MidFT(6,parname[b].Len());
				partype[b]=2;
				fptable[nf][b]=2;
				vt=1;
			}
			if(parname[b].Cmnd("REG"))
			{
				// Pass parameter directly to a register
				vt=1;
				parname[b]=parname[b].MidFT(3,parname[b].Len());
				if(parname[b].Len()==0)
				{
					// It must be an 8 bit register...
					switch(parname[b].txt[0])
					{
					case 'B':
						c=5;
						break;
					case 'C':
						c=6;
						break;
					case 'D':
						c=7;
						break;
					case 'E':
						c=8;
						break;
					case 'H':
						c=9;
						break;
					case 'L':
						c=10;
						break;
					default:
						cout << "\nInvalid register parameter. Valid registers are: B, C, D, E, H, L, BC, DE, and HL.";
						EndProgram();
					}
					// Assigns parameter type according to register id
					partype[b]=c;
					fptable[nf][b]=c;
				}
				else
				{
					// It must be a 16 bit register...
					int vr=0;
					if(parname[b].Cmnd("BC"))
					{
						vr=1;
						partype[b]=11;
						fptable[nf][b]=11;
					}
					if(parname[b].Cmnd("DE"))
					{
						vr=1;
						partype[b]=12;
						fptable[nf][b]=12;
					}
					if(parname[b].Cmnd("HL"))
					{
						vr=1;
						partype[b]=13;
						fptable[nf][b]=13;
					}
					if((vr==0)||(parname[b].Len()>1))
					{
						cout << "\nInvalid register parameter. Valid registers are: B, C, D, E, H, L, BC, DE, and HL.";
						EndProgram();	
					}
				}
			}
			if(vt==0)
			{
				cout << "\nInvalid parameter type in function declaration: ";
				M_CF;
				parname[b].Disp();
				EndProgram();
			}
		}
		ipars[nf]=ip;
		while(cl!="{")
			M_GL;
		// Got rid of unnecessary MEMRESETs and a {
		return 1;
	}
	return 0;
}

int GetVarType(ZString s)
{
	// Finds out if a given parameter is single or double etc.
	if(s.Cmnd("MEM")==1)
		return 2;
	if(s.Cmnd("PARV")==1)
		return partype[s.txt[4]-48];
	// If it's not a temp or ret mem variable, scan the variable
	//   symbolic table to find the matching variable
	for(a=1;a<=vtotal;a++)
		if(var[a]==s)return vtype[a];
	// Since it's not a variable, it must be a number, so convert it
	b=atoi(s.txt);
	char ft[20];
	itoa(b,ft,10);
	if(strcmp(ft,s.txt)!=0)
	{
		cout << "\nUndeclared variable found.";
		M_CF;
		s.Disp(3);
		EndProgram();
	}
	if(b<256)return 3;
	return 4;
	// Returns 3 for single number and 4 for double number
}

void MakeEqual()
{
	// Makes equality
	a=cl.Find("=");
	p1=cl.MidFT(0,a-1);
	p2=cl.MidFT(a+1,cl.Len());
	int d=GetVarType(p1);
	c=GetVarType(p2);
	b=d;
	// Got the names of the operands
	if((b==1)&&(c==1))
	{
		// Load 1 byte into 1 byte
		temp = ZSCon("@ld A, (") + p2 + ")";M_PLT;
		temp = ZSCon("@ld (") + p1 + "), A";M_PLT;
	}
	if((b==2)&&(c==2))
	{
		// Load 2 byte into 2 byte
		temp = ZSCon("@ld HL, (") + p2 + ")";M_PLT;
		temp = ZSCon("@ld (") + p1 + "), HL";M_PLT;
	}
	if((b==2)&&(c==1))
	{
		// Load 1 byte into 2 byte
		temp = ZSCon("@ld A, (") + p2 + ")";M_PLT;
		temp = "@ld C, A";M_PLT;
		temp = "@ld B, 0";M_PLT;
		temp = ZSCon("@ld (") + p1 + "), BC";M_PLT;
	}
	if((b==1)&&(c==2))
	{
		if((p2.Cmnd("MEM"))||(p2.Cmnd("TEMPM")))
		{
			// Load lower byte of mem or temp variable into single variable
			temp = ZSCon("@ld A, (") + p2 + ")";M_PLT;
			temp = ZSCon("@ld (") + p1 + "), A";M_PLT;
		}
		else
		{
			cout << "\nError trying to load double variable into single.";
			M_CF;
			p1.Disp(3);
			p2.Disp(2);
			EndProgram();
		}
	}
	if((b==1)&&(c==3))
	{
		// Load 1 byte number into 1 byte variable
		temp = ZSCon("@ld A,") + p2;M_PLT;
		temp = ZSCon("@ld (") + p1 + "), A";M_PLT;
	}
	if((b==2)&&(c==4))
	{
		// Load 2 byte number into 2 byte variable
		temp = ZSCon("@ld HL,") + p2;M_PLT;
		temp = ZSCon("@ld (") + p1 + "), HL";M_PLT;
	}
	if((b==2)&&(c==3))
	{
		// Load 1 byte number into 2 byte variable
		temp = ZSCon("@ld C, ") + p2;M_PLT;
		temp  = "@ld B, 0";M_PLT;
		temp = ZSCon("@ld (") + p1 + "),  BC";M_PLT;
	}
	if((b==1)&&(c==4))
	{
		cout << "\nError trying to load large number into single variable.";
		EndProgram();
	}
	nw=1;
}

ZString Convert(ZString s, int ov)
{
	int ts = GetVarType(s);
	ZString opv;
	opv=s;
	if(ts!=2)opv=ZSCon("OPVAR") + ov;
	if(ts==1)
	{
		// Convert single variable to double variable
		temp = "@XOR A";M_PLT;
		temp = "@ld B, A";M_PLT;
		temp = ZSCon("@ld A, (") + s + ")";M_PLT;
		temp = "@ld C, A";M_PLT;
		temp = ZSCon("@ld (") + opv + "), BC";M_PLT;
	}
	if(ts==3)
	{
		// Convert single number to double variable
		temp = "@XOR A";M_PLT;
		temp = "@ld B, A";M_PLT;
		temp = ZSCon("@ld A, ") + s;M_PLT;
		temp = "@ld C, A";M_PLT;
		temp = ZSCon("@ld (") + opv + "), BC";M_PLT;
	}
	if(ts==4)
	{
		// Convert double number to double variable
		temp = ZSCon("@ld BC, ") + s;M_PLT;
		temp = ZSCon("@ld (") + opv + "), BC";M_PLT;
	}
	return opv;
}

void CallFunction()
{
	// Places code for calling a function
	ZString parn[10];
	int i = GetFunPars(cl,parn);
	a=cl.Find("(");
	ZString fn = cl.MidFT(0,a-1);
	int cc=0;
	for(int aa=0;aa<=nf;aa++)
	{
		// Searches for the function in the function table
		if(fntable[aa]==fn)
		{
			cc=1;
			if(ipars[aa]!=i)
			{
				cout << "\nInvalid number of parameters.\n" << i << " " << ipars[aa];
				EndProgram();
			}
			// Ensured that number of parameters passed was correct
			//   and now function id number is in aa
			break;
		}
	}
	if(cc==0)
	{
		cout << "\nUndeclared function found.";
		M_CF;
		fn.Disp(2);
		EndProgram();
	}
	fcalled[aa]=1;
	if(fn=="GOTO")
	{
		// Processes "GOTO" keyword
		temp=ZSCon("@JUMP_(") + parn[1] + ")";M_PLT;
		return;
	}
	if(fn=="PRINT")
	{
		// Processes "PRINT" keyword
		if(parn[3].Cmnd("USTR"))
		{
			// Printing text
			temp=parn[3].MidFT(4,parn[3].Len());
			b=atoi(temp.txt);
			temp = ZSCon("PRINTT(") + parn[1] + ZSCon(",") + parn[2] + ZSCon(",") + parn[3] + ZSCon(",") + ustrlen[b] + ")";M_PLT;
		}
		else
		{
			// Printing a number
			temp = ZSCon("PRINTN") + cl.MidFT(5,cl.Len());M_PLT;
		}
		return;
	}
	if(fn=="RETURN")
	{
		// Processes "RETURN" keyword
		retexist=1;
		if(crt==1)
			// 1 byte return
			temp = ZSCon("RETV1=")+parn[1];
		if(crt==2)
			// 2 byte return
			temp = ZSCon("RETV2=")+parn[1];
		if(crt==0)
		{
			cout << "\nA void function cannot return a value.";
			EndProgram();
		}
		if(crt!=0)M_PLT;
		PutLine(outfile,"@ret");
		return;
	}
	if(fn=="LOAD")
	{
		// Returns value stored in permanent variable
		temp = ZSCon("@ld DE, ") + parn[1];M_PLT;
		temp = "@ld HL, (PROGRAM_ADDR)";M_PLT;
		temp = "@ADD HL, DE";M_PLT;
		temp = "@ld A, (HL)";M_PLT;
		temp = "@ld (RETV2), A";M_PLT;
		temp = "@INC HL";M_PLT;
		temp = "@ld A, (HL)";M_PLT;
		temp = "@ld (RETV2+1), A";M_PLT;
		return;
	}
	if(fn=="SAVE")
	{
		// Saves something in a permanent variable
		parn[1]=Convert(parn[1],1);	
		parn[1] = ZSCon("(") + parn[1] + ")";
		temp = "@ld DE, (PROGRAM_ADDR)";M_PLT;
		temp = ZSCon("@ld HL, ") + parn[2];M_PLT;
		temp = "@ADD HL, DE";M_PLT;
		temp = ZSCon("@ld A, ") + parn[1];M_PLT;
		temp = "@ld (HL), A";M_PLT;
		parn[1] = parn[1].MidFT(0,parn[1].Len()-1) + "+1)";
		temp = "@INC HL";M_PLT;
		temp = ZSCon("@ld A, ") + parn[1];M_PLT;
		temp = "@ld (HL), A";M_PLT;
		return;
	}
	if(fn=="FOR")
	{
		// Processes "FOR" keyword
		// Gets rid of unnecessary {
		while(cl!="{")
			M_GL;
		// Variable = starting value
		temp = parn[1] + ZSCon("=") + parn[2] + ZSCon("-") + parn[4];M_PLT;
		cn++;
		lcn=cn;
		nesttype[cn]=2;
		tl++;
		nestgoto[cn] = ZSCon("NESTLBL") + tl;
		temp = nestgoto[cn] + ":";M_PLT;
		// Increase variable by STEP (last parameter)
		temp = parn[1] + ZSCon("=") + parn[1] + ZSCon("+") + parn[4];M_PLT;
		// Check to see if variable = ending value 
		temp = ZSCon("IF(") + parn[1] + ZSCon(">(") + parn[3] + "+1))";M_PLT;
		temp = "{";M_PLT;
		tl++;
		nestlabel[cn] = ZSCon("NESTLBL") + tl;
		temp = ZSCon("@JUMP_(") + nestlabel[cn] + ")";M_PLT;
		temp = "}";M_PLT;
		return;
	}
	if(fn=="BYTEARRAY")
	{
		// Process byte array data
		// First get rid of irrelevant {
		while(cl!="{")
			M_GL;
		partype[2]=GetVarType(parn[2]);
		partype[3]=GetVarType(parn[3]);
		if((partype[2]!=3)||(partype[3]!=3))
		{
			cout << "\nBYTEARRAY expects two single numbers as the last 2 parameters.";
			EndProgram();
		}
		nars++;
		arname[nars]=parn[1];
		aorl[nars]=1;
		arx[nars]=atoi(parn[2].txt);
		temp=arx[nars];
		ary[nars]=atoi(parn[3].txt);
		tl++;
		temp = ZSCon("@jr NESTLBL") + tl;M_PLT;
		temp = parn[1] + ":";M_PLT;
		temp=ZSCon("DOUBLE")+parn[1];
		AssignVar(temp,2);
		vtype[vtotal]=4;
		ilv++;
		lv[ilv]=vtotal;
		while(1)
		{
			// Keep reading data until } is reached
			M_GL;
			if(cl=="}")break;
			if((cl!="MEMRESET")&&(cl.txt[0]!='\''))
			{
				// Only get data if it really is data (and not a comment or MEMRESET)
				temp = ZSCon("@.db ") + cl;M_PLT;
			}
		}	
		temp = ZSCon("NESTLBL") + ZSCon(tl) + ":";M_PLT;
		nw=1;
		return;
	}
	if(fn=="LIST")
	{
		// Process list data
		// First get rid of irrelevant {
		while(cl!="{")
			M_GL;
		tl++;
		temp = ZSCon("@jr NESTLBL") + tl;M_PLT;
		temp = parn[1] + ":";M_PLT;
		temp=ZSCon("DOUBLE")+parn[1];
		AssignVar(temp,2);
		vtype[vtotal]=4;
		ilv++;
		lv[ilv]=vtotal;
		nars++;
		arname[nars]=parn[1];
		aorl[nars]=2;
		while(1)
		{
			// Keep reading data until } is reached
			M_GL;
			if(cl=="}")break;
			if((cl!="MEMRESET")&&(cl.txt[0]!='\''))
			{
				// Only get data if it really is data (and not a comment or MEMRESET)
				temp = ZSCon("@.db ") + cl;M_PLT;
			}
		}	
		temp = ZSCon("NESTLBL") + ZSCon(tl) + ":";M_PLT;
		nw=1;
		return;
	}
	if(fn=="LGET")
	{
		// Get a value from a list
		for(int aa=1;aa<=nars;aa++)
			// Find list ID number
			if(arname[aa]==parn[1])break;
		if(arname[aa]!=parn[1])
		{
			parn[1].Disp(1);
			cout << "\nUndeclared list.";
			EndProgram();
		}
		if(aorl[aa]!=2)
		{
			parn[1].Disp(1);
			cout << "\nArray found where a list is expected (LGET).";
			EndProgram();
		}
		// Homogenize single variable and single number parameter types
		if(GetVarType(parn[2])==1)
			parn[2] = ZSCon("(") + parn[2] + ")";
		temp = ZSCon("@ld A, ") + parn[2];M_PLT;
		temp = "@DEC A";M_PLT;
		parn[2]="(ASCII2)";
		// The ASCIIs are only used as temporary variables (nothing to do with ASCII)
		temp = ZSCon("@ld ") + parn[2] + ", A";M_PLT;
		temp = "@XOR A";M_PLT;
		temp = "@ld (ASCII2+1), A";M_PLT;
		temp = "@ld HL, (PROGRAM_ADDR)";M_PLT;
		temp = ZSCon("@ld DE, ") + parn[1];M_PLT;
		temp = "@ADD HL, DE";M_PLT;
		temp = "@ld DE, (ASCII2)";M_PLT;
		temp = "@ADD HL, DE";M_PLT;
		temp = "@ld A, (HL)";M_PLT;
		temp = "@ld (RETV1), A";M_PLT;
		return;
	}
	if(fn=="LPUT")
	{
		// Put a value in a list
		chksum=1;
		for(int aa=1;aa<=nars;aa++)
			// Find list ID number
			if(arname[aa]==parn[1])break;
		if(arname[aa]!=parn[1])
		{
			parn[1].Disp(1);
			cout << "\nUndeclared list.";
			EndProgram();
		}
		if(aorl[aa]!=2)
		{
			parn[1].Disp(1);
			cout << "\nArray found where a list is expected (LPUT).";
			EndProgram();
		}
		// Homogenize single variable and single number parameter types
		if(GetVarType(parn[2])==1)
			parn[2] = ZSCon("(") + parn[2] + ")";
		if(GetVarType(parn[3])==1)
			parn[3] = ZSCon("(") + parn[3] + ")";
		temp = ZSCon("@ld A, ") + parn[2];M_PLT;
		temp = "@DEC A";M_PLT;
		parn[2]="(ASCII2)";
		// The ASCIIs are only used as temporary variables (nothing to do with ASCII)
		temp = ZSCon("@ld ") + parn[2] + ", A";M_PLT;
		temp = "@XOR A";M_PLT;
		temp = "@ld (ASCII2+1), A";M_PLT;
		temp = "@ld HL, (PROGRAM_ADDR)";M_PLT;
		temp = ZSCon("@ld DE, ") + parn[1];M_PLT;
		temp = "@ADD HL, DE";M_PLT;
		temp = "@ld DE, (ASCII2)";M_PLT;
		temp = "@ADD HL, DE";M_PLT;
		temp = ZSCon("@ld A, ") + parn[3];M_PLT;
		temp = "@ld (HL), A";M_PLT;
		return;
	}
	if(fn=="SPRITE")
	{	
			// Process bit array data
			// First get rid of irrelevant {
			while(cl!="{")
				M_GL;
			tl++;
			temp = ZSCon("@jr NESTLBL") + tl;M_PLT;
			temp = parn[1] + ":";M_PLT;
			temp=ZSCon("DOUBLE")+parn[1];
			AssignVar(temp,2);
			vtype[vtotal]=4;
			ilv++;
			lv[ilv]=vtotal;
			temp = "@.db 8,8";M_PLT;
			int aa=0;
			while(1)
			{
				// Keep reading data until } is reached
				M_GL;
				if(cl=="}")break;
				if((cl.txt[0]=='1')||(cl.txt[0]=='0'))
				{
					aa++;
					// Only get data if it really is data (and not a comment or MEMRESET)
					temp = ZSCon("@.db %") + cl;M_PLT;
				}
			}	
			if(aa!=8)
			{
				cout << "\nSPRITE expects eight rows of data.";
				EndProgram();
			}
			temp = ZSCon("NESTLBL") + ZSCon(tl) + ":";M_PLT;
			nw=1;
			return;
	}
	if(fn=="WHILE")
	{
		// Processes "while" at end of a do-while loop
		temp = ZSCon("IF(1==") + parn[1] + ")";M_PLT;
		temp = "{";M_PLT;
		temp = ZSCon("GOTO(") + nestlabel[cn] + ")";M_PLT;
		temp = "}";M_PLT;
		cn--;
		return;
	}
	if(fn=="ARPUT")
	{
		// Put a value in an array
		chksum=1;
		for(int aa=1;aa<=nars;aa++)
			// Find array ID number
			if(arname[aa]==parn[1])break;
		if(arname[aa]!=parn[1])
		{
			parn[1].Disp(1);
			cout << "\nUndeclared array.";
			EndProgram();
		}
		if(aorl[aa]!=1)
		{
			parn[1].Disp(1);
			cout << "\nList found where an array is expected (ARPUT).";
			EndProgram();
		}
		// Homogenize single variable and single number parameter types
		if(GetVarType(parn[2])==1)
			parn[2] = ZSCon("(") + parn[2] + ")";
		if(GetVarType(parn[3])==1)
			parn[3] = ZSCon("(") + parn[3] + ")";
		if(GetVarType(parn[4])==1)
			parn[4] = ZSCon("(") + parn[4] + ")";
		temp = ZSCon("@ld A, ") + parn[2];M_PLT;
		temp = "@DEC A";M_PLT;
		parn[2]="(ASCII2)";
		// The ASCIIs are only used as temporary variables (nothing to do with ASCII)
		temp = ZSCon("@ld ") + parn[2] + ", A";M_PLT;
		temp = ZSCon("@ld A, ") + parn[3];M_PLT;
		temp = "@DEC A";M_PLT;
		parn[3]="(ASCII3)";
		temp = ZSCon("@ld ") + parn[3] + ", A";M_PLT;
		parn[2]="ASCII2";
		parn[3]="ASCII3";
		temp = "@XOR A";M_PLT;
		temp = "@ld (ASCII2+1), A";M_PLT;
		temp = "@ld (ASCII3+1), A";M_PLT;
		temp = ZSCon("ASCII=") + parn[3] + ZSCon("*") + arx[aa] + ZSCon("+") + parn[2];M_PLT;
		temp = "@ld HL, (PROGRAM_ADDR)";M_PLT;
		temp = ZSCon("@ld DE, ") + parn[1];M_PLT;
		temp = "@ADD HL, DE";M_PLT;
		temp = "@ld DE, (ASCII)";M_PLT;
		temp = "@ADD HL, DE";M_PLT;
		temp = ZSCon("@ld A, ") + parn[4];M_PLT;
		temp = "@ld (HL), A";M_PLT;
		return;
	}
	if(fn=="ARGET")
	{
		// Get a value from an array
		for(int aa=1;aa<=nars;aa++)
			// Find array ID number
			if(arname[aa]==parn[1])break;
		if(arname[aa]!=parn[1])
		{
			parn[1].Disp(1);
			cout << "\nUndeclared array.";
			EndProgram();
		}
		if(aorl[aa]!=1)
		{
			parn[1].Disp(1);
			cout << "\nList found where an array is expected (ARGET).";
			EndProgram();
		}
		// Homogenize single variable and single number parameter types
		if(GetVarType(parn[2])==1)
			parn[2] = ZSCon("(") + parn[2] + ")";
		if(GetVarType(parn[3])==1)
			parn[3] = ZSCon("(") + parn[3] + ")";
		temp = ZSCon("@ld A, ") + parn[2];M_PLT;
		temp = "@DEC A";M_PLT;
		parn[2]="(ASCII2)";
		// The ASCIIs are only used as temporary variables (nothing to do with ASCII)
		temp = ZSCon("@ld ") + parn[2] + ", A";M_PLT;
		temp = ZSCon("@ld A, ") + parn[3];M_PLT;
		temp = "@DEC A";M_PLT;
		parn[3]="(ASCII3)";
		temp = ZSCon("@ld ") + parn[3] + ", A";M_PLT;
		parn[2]="ASCII2";
		parn[3]="ASCII3";
		temp = "@XOR A";M_PLT;
		temp = "@ld (ASCII2+1), A";M_PLT;
		temp = "@ld (ASCII3+1), A";M_PLT;
		temp = ZSCon("ASCII=") + arx[aa] + ZSCon("*") + parn[3] + ZSCon("+") + parn[2];M_PLT;
		temp = "@ld HL, (PROGRAM_ADDR)";M_PLT;
		temp = ZSCon("@ld DE, ") + parn[1];M_PLT;
		temp = "@ADD HL, DE";M_PLT;
		temp = "@ld DE, (ASCII)";M_PLT;
		temp = "@ADD HL, DE";M_PLT;
		temp = "@ld A, (HL)";M_PLT;
		temp = "@ld (RETV1), A";M_PLT;
		return;
	}
	if(fn=="IF")
	{
		// Processes "IF" keyword
		// Get rid of unnecessary {
		while(cl!="{")
			M_GL;
		lc++; // Line counter correction
		cn++;
		nesttype[cn]=4;
		a=GetVarType(parn[1]);
		if(a==3)
			// Just in case sombody actually typed "1" as a parameter...
			temp = ZSCon("@ld A,") + parn[1];
		else
			// Always assumed to be a double variable (a MEM one practically)
			//   but with only the lower bit significant
			temp = ZSCon("@ld A,(") + parn[1] + ")";M_PLT;
		temp = "@cp 1";M_PLT;
		tl++;
		temp = ZSCon("@jr Z, NESTLBL") + tl;M_PLT;
		tl++;
		nestlabel[cn] = ZSCon("NESTLBL") + tl;
		temp = ZSCon("@JUMP_(") + nestlabel[cn] + ")";M_PLT;
		a=tl-1;
		temp = ZSCon("@NESTLBL") + ZSCon(a) + ":";M_PLT;
		return;
	}
	// By now, we know the function must be a user declared function
	if(inf==1)
	{
		// Checks to see if the call was made within a function
		//   declaration and if so saves old parameters
		for(a=1;a<=ipars[nf];a++)
		{
			// Goes through "upper" function's parameters and
			//   saves them in the register stack
			if(partype[a]==1)
			{
				// 1 byte parameter saved in A
				temp=ZSCon("@ld A, (PARV") + ZSCon(a) + ")";M_PLT;
				temp="@PUSH AF";M_PLT;
			}
			if(partype[a]==2)
			{
				// 2 byte parameter saved in HL
				temp=ZSCon("@ld HL, (PARV") + ZSCon(a) + ")";M_PLT;
				temp = "@PUSH HL";M_PLT;
			}
			// Note that parameters that are registers are NOT saved
		}
	}
	for(int bb=1;bb<=i;bb++)
	{
		// Converts all passed parameters to PARVx
		a=GetVarType(parn[bb]);
		if(fptable[aa][bb]==1)
		{
			// Function parameter is single
			if(a==1)
			{
				// Single passed to single
				temp=ZSCon("@ld A, (") + parn[bb] + ")";M_PLT;
				temp=ZSCon("@ld (PARV") + ZSCon(bb) + "), A";M_PLT;
			}
			if(a==3)
			{
				// Single number passed to single
				temp=ZSCon("@ld A, ") + parn[bb];M_PLT;
				temp=ZSCon("@ld (PARV") + ZSCon(bb) + "), A";M_PLT;
			}
			if(a==4)
			{
				cout << "\nCannot pass double number to single parameter.";
				EndProgram();
			}
			if(a==2)
			{
				// Double passed to single (lower bits taken)
				temp=ZSCon("@ld A, (") + parn[bb] + ")";M_PLT;
				temp=ZSCon("@ld (PARV") + ZSCon(bb) + "), A";M_PLT;
			}
		}
		if(fptable[aa][bb]==2)
		{
			// Function parameter is double
			if(a==1)
			{
				// Single passed to double
				temp = "@PUSH BC";M_PLT;
				temp = ZSCon("@ld A, (") + parn[bb] + ")";M_PLT;
				temp = "@ld C, A";M_PLT;
				temp = "@ld B, 0";M_PLT;
				temp = ZSCon("@ld (PARV") + ZSCon(bb) + "), BC";M_PLT;
				temp = "@POP BC";M_PLT;
			}
			if(a==2)
			{
				// Double passed to double
				temp = ZSCon("@ld HL, (") + parn[bb] + ")";M_PLT;
				temp = ZSCon("@ld (PARV") + ZSCon(bb) + "), HL";M_PLT;
			}
			if(a==3)
			{
				// Single number passed to double
				temp = "@PUSH BC";M_PLT;
				temp = ZSCon("@ld C, ") + parn[bb];M_PLT;
				temp  = "@ld B, 0";M_PLT;
				temp = ZSCon("@ld (PARV") + ZSCon(bb) + "),  BC";M_PLT;
				temp = "@POP BC";M_PLT;
			}
			if(a==4)
			{
				// Double number passed to double
				temp = ZSCon("@ld HL,") + parn[bb];M_PLT;
				temp = ZSCon("@ld (PARV") + ZSCon(bb) + "), HL";M_PLT;
			}
		}
		if(fptable[aa][bb]>2)
		{
			if(fptable[aa][bb]<11)
			{
				// Function parameter is an 8 bit register
				if(a==1)
				{
					// Single variable passed to 8 bit register
					temp = ZSCon("@ld A, (") + parn[bb] + ")";M_PLT;
					temp = ZSCon("@ld ") + reg[fptable[aa][bb]] + ", A";M_PLT;
				}
				if(a==2)
				{
					// Double variable passed to 8 bit register
					temp = ZSCon("@ld A, (") + parn[bb] + ")";M_PLT;
					temp = ZSCon("@ld ") + reg[fptable[aa][bb]] + ", A";M_PLT;
				}
				if(a==3)
				{
					// Single number passed to 8 bit register
					temp = ZSCon("@ld ") + reg[fptable[aa][bb]] + ZSCon(", ") + parn[bb];M_PLT;
				}
				if(a==4)
				{
					cout << "\nCannot pass double number to single register parameter.";
					EndProgram();
				}
			}
			else
			{
				// Function parameter is a 16 bit register
				if(a==1)
				{
					// Single variable passed to 16 bit register
					temp = "@PUSH BC";M_PLT;
					temp = "@ld B, 0";M_PLT;
					temp = ZSCon("@ld A, (") + parn[bb] + ")";M_PLT;
					temp="@ld C, A";M_PLT;
					temp = "@ld (TEMPM), BC";M_PLT;
					temp=ZSCon("@ld ") + reg[fptable[aa][bb]] + ", (TEMPM)";M_PLT;
					temp = "@POP BC";M_PLT;
				}
				if(a==2)
				{
					// Double variable passed to 16 bit register
					temp = ZSCon("@ld ") + reg[fptable[aa][bb]] + ZSCon(", (") + parn[bb] + ")";M_PLT;
				}
				if(a==3)
				{
					// Single number passed to 16 bit register
					temp = "@PUSH BC";M_PLT;
					temp = "@ld B, 0";M_PLT;
					temp = ZSCon("@ld C, ") + parn[bb];M_PLT;
					temp = "@ld (TEMPM), BC";M_PLT;
					temp = ZSCon("@ld ") + reg[fptable[aa][bb]] + ", (TEMPM)";M_PLT;
					temp = "@POP BC";M_PLT;
				}
				if(a==4)
				{
					// Double number passed to 16 bit register
					temp = ZSCon("@ld ") + reg[fptable[aa][bb]] + ZSCon(", ") + parn[bb];M_PLT;
				}
			}
		}
	}
	temp = ZSCon("@CALL_(FSTART") + ZSCon(aa) + ")";M_PLT;
	if(inf==1)
	{
		// Checks if function call was made inside a function
		//   declaration and if so loads old parameters
		for(a=ipars[nf];a>=1;a--)
		{
			// Goes through "upper" function's parameters and
			//   loads them from the register stack (in reverse order)
			if(partype[a]==1)
			{
				// 1 byte parameter loaded from A
				temp = "@POP AF";M_PLT;
				temp = ZSCon("@ld (PARV") + ZSCon(a) + "), A";M_PLT;
			}
			if(partype[a]==2)
			{
				// 2 byte parameter loaded from HL
				temp = "@POP HL";M_PLT;
				temp = ZSCon("@ld (PARV") + ZSCon(a) + "), HL";M_PLT;
			}
			// Note that parameters that were registers were NOT saved and thus are
			//   not loaded
		}
	}
}

void AsmEqualityCheck(char* jtype)
{
		temp = ZSCon("@ld HL, (") + p1 +")";M_PLT;
		temp = ZSCon("@ld DE, (") + p2 + ")";M_PLT;
		temp = "@ld BC, 0";M_PLT;
		temp = "@ld (RETV2), BC";M_PLT;
		temp = "@OR A \\ SBC HL, DE";M_PLT;
		tl++;
		temp = ZSCon("@jr ") + jtype + ZSCon(", NESTLBL") + tl;M_PLT;
		temp = "@ld A, 1";M_PLT;
		temp = "@ld (RETV2), A";M_PLT;
		temp = ZSCon("NESTLBL") + ZSCon(tl) + ":";M_PLT;
}

void ProcessAri()
{
	int bb;
	for(int a=0;a<10;a++)
	{
		// Finds out what operator is present and where it is
		bb=cl.Find(cl.oplist[a]);
		if(bb!=-1)break;
	}
	p1=cl.MidFT(0,bb-1);
	if((a<2)||(a>7))
		p2=cl.MidFT(bb+2,cl.Len());
	else
		p2=cl.MidFT(bb+1,cl.Len());
	// Now, convert everything so that both parameters are double
	//   variables (the most complex combo possible)
	p1=Convert(p1,1);
	p2=Convert(p2,2);
	// Check which operation it is...
	switch(a)
	{
	case 0: // == (equality check)
		AsmEqualityCheck("nz");
		break;
	case 1: // <> (not-equal-to check)
		AsmEqualityCheck("z");
		break;
	case 2: // < (less than check)
		AsmEqualityCheck("nc");
		break;
	case 3: // > (greater than check)
		AsmEqualityCheck("c");
		break;
	case 4: // * (multiplication)
		temp = "TEMPM=0";M_PLT;
		temp = ZSCon("@ld A, (") + p1 + ")";M_PLT;
		temp = "@ld B, A";M_PLT;
		tl++;
		temp = ZSCon("NESTLBL") + tl + ZSCon(":");M_PLT;
		temp = "@PUSH BC";M_PLT;
		temp = ZSCon("TEMPM=TEMPM+") + p2;M_PLT;
		temp = "@POP BC";M_PLT;
		temp = ZSCon("@djnz NESTLBL") + tl;M_PLT;
		temp = "RETV2=TEMPM";M_PLT;
		break;
	case 5: // / (division)
		temp = "TEMPM=0";M_PLT;
		tl++;
		temp = ZSCon("NESTLBL") + tl + ZSCon(":");M_PLT;
		temp = "@ld A, (TEMPM)";M_PLT;
		temp = "@INC A";M_PLT;
		temp = "@ld (TEMPM), A";M_PLT;
		temp = p1 + ZSCon("=") + p1 + ZSCon("-") + p2;M_PLT;		
		temp = ZSCon("@jr nc, NESTLBL") + tl;M_PLT;
		temp = "RETV2=TEMPM-1";M_PLT;
		break;
	case 6: // + (addition)
		temp = ZSCon("@ld HL, (") + p1 + ")";M_PLT;
		temp = ZSCon("@ld BC, (") + p2 + ")";M_PLT;
		temp = "@ADD HL, BC";M_PLT;
		temp = "@ld (RETV2), HL";M_PLT;
		break;
	case 7: // - (subtraction)
		temp = ZSCon("@ld HL, (") + p1 +")";M_PLT;
		temp = ZSCon("@ld DE, (") + p2 + ")";M_PLT;
		temp = "@OR A \\ SBC HL, DE";M_PLT;
		temp = "@ld (RETV2), HL";M_PLT;
		break;
	case 8: // && (logical AND)
		temp = "RETV2=0";M_PLT;
		temp = ZSCon("IF(") + p1 +")";M_PLT;
		temp = "{";M_PLT;
		temp = ZSCon("IF(") + p2 + ")";M_PLT;
		temp = "{";M_PLT;
		temp = "RETV2=1";M_PLT;
		temp = "}";M_PLT;
		temp = "}";M_PLT;
		break;
	case 9: // || (logical OR)
		temp = "RETV2=0";M_PLT;
		temp = ZSCon("IF(") + p1 + ")";M_PLT;
		temp = "{";M_PLT;
		temp = "RETV2=1";M_PLT;
		temp = "}";M_PLT;
		temp = ZSCon("IF(") + p2 + ")";M_PLT;
		temp = "{";M_PLT;
		temp = "RETV2=1";M_PLT;
		temp = "}";M_PLT;
	}

}

int IntegrityCheck(ZString s)
{
	// Check to see if any parentheses are missing in a command line
	b=0;
	pc=0;
	int pcp = 0;
	for(a=0;a<=s.Len();a++)
	{
		// Increases and decreases p-count appropriatly
		c=s.txt[a];
		if(c=='(')
		{
			b++;
			pc++;
			pcp++;
		}
		if(c==')')b--;
	}
	if(b<0)
	{
		cout << "\nMissing opening parentheses";
		EndProgram();
	}
	if(b>0)
	{
		cout << "\nMissing closing parentheses";
		EndProgram();
	}
	return pcp;
}

void ArrayBrackets()
{
	// Translate array brackets into ARPUT or ARGET functions
	// or list arrays to LGET and LPUT functions
	int aa=cl.Find("[");
	int bb=cl.SLeftForOp(aa);
	int cc=cl.Find("]");
	temp = cl.MidFT(aa,cc);
	if(bb==-1)
	{
		// There are no operators to the left, so it's either an
		// ARPUT or LPUT
		if(temp.Find(",")==-1)
		{
			//Translate to LPUT (no comma found)
			temp = cl.MidFT(0,aa-1);
			temp = ZSCon("LPUT(") + temp + ZSCon(",") + cl.MidFT(aa+1,cc-1) + ZSCon(",") + cl.MidFT(cc+2,cl.Len()) + ")";
		}
		else
		{
			// Tranlate to ARPUT
			temp=cl.MidFT(0,aa-1);
			temp = ZSCon("ARPUT(") + temp + ZSCon(",") + cl.MidFT(aa+1,cc-1) + ZSCon(",") + cl.MidFT(cc+2,cl.Len()) + ")";
		}
		M_PLT;
	}
	else
	{
		// It's either an ARGET or LGET
		if(temp.Find(",")==-1)
		{
			// Translate to LGET (no comma found)
			temp = cl.MidFT(bb+1,aa-1);
			temp = ZSCon("MEM9=LGET(") + temp + ZSCon(",") + cl.MidFT(aa+1,cc-1) + ")";
		}
		else
		{
			// Translate to ARGET
			temp = cl.MidFT(bb+1,aa-1);
			temp = ZSCon("MEM9=ARGET(") + temp + ZSCon(",") + cl.MidFT(aa+1,cc-1) + ")";
		}
		M_PLT;
		temp = cl.MidFT(0,bb) + ZSCon("MEM9") + cl.MidFT(cc+1,cl.Len());M_PLT;
	}
	return;
}

void Parser(ifstream infile, ofstream outfile)
{
	int deb=0;
	int pcount;
	while(nw==1)
	{
		// Go through the work files as long as there 
		// is more to be parsed
		M_FLIP;
		nw=0;
		lc=0;
		cout << "*";
		M_CF;
		while(infile)
		{
			// Go through the file line by line with parser
			M_GL;
			if(_kbhit())
			{
				// User keystroke processing
				cout << "\nCompilation paused. Press escape to quit, any other key to continue.";
				M_CF;
				getch();
				a=getch();
				if(a==27)
				{
					cout << "\nCompilation terminated by user.";
					parsing=0;
					preparsing=0;
					M_CF;
					EndProgram();
				}
				cout << "\nCompilation resumed.\nParsing ";
				M_CF;
			}
			// Ignores stray blank lines
			if(!isgraph(cl.txt[0]))goto readnext;
			if((cl.txt[0]=='@')||(cl.txt[0]=='\'')||(cl.txt[cl.Len()]==':'))
			{
				// If the line stars with '' then mark it as the 
				// error-marker original source code line
				if((cl.txt[0]=='\'')&&(cl.txt[1]=='\''))
					cline = cl.MidFT(2,cl.Len());
				// Copy direct assembles lines, comments  and 
				//   labels directly into outfile
				M_PLC;
				goto readnext;
			}
			// Check if entering or exiting a function declaration
			//   that was already parsed once.
			if(cl.Cmnd("FSTART"))inf=1;
			if(cl.Cmnd("FEND"))inf=0;
			if(inf==1)
			{
				// Inside a function declaration
				for(int aa=1;aa<=ip;aa++)
				{
					// Replaces parameter names with PARVx 
					if(fptable[nf][aa]<3)
					{
						// Only does so if parameter is not a register
						p1=parname[aa];
						p2 = ZSCon("PARV") + aa;
						if(cl.Find(p1)!=-1)cl.Replace(p1,p2);
					}
				}
			}
			pcount = IntegrityCheck(cl);
			if(cl==ZSCon("MEMRESET"))
			{
				// Clear internal temp mem counter (Practically add 1 
				// to the current line counter)
				// This code is located here to improve compilation
				// speed (it should theoretically be in keyword 
				// processing)
				lc++;
				M_PLC;
				goto readnext;
			}
			if(debug==true)cl.Disp(2);
			a=cl.Find("]");
			if(a!=-1)
			{
				// Array brackets found
				ArrayBrackets();
				goto readnext;		
			}
			a=ProcessFunDecs();
			if(a==1)goto readnext;
			// Checks for function declarations and processes them
			//   if they exist. 
			a=cl.Flag();
			// Finds out what type of "stuff" is in the command line
			if(cl.IsFunc()==1)
			{
				// Function to be "pushed up" perhaps
				if(cl.IsPar()==1)
				{
					// First check if there aren't any stray parentheses
					//   and if so, push them up
					PushUpPars();
					goto readnext;
				}
				if(PushUpFunc()==1)goto readnext;
				if(cl.IsAri()==1)
				{
					// First check if there aren't any stray arithmetic
					//   operations, and if so, push them up
					PushUpOps();
					goto readnext;
				}
				if(cl.IsEqual()==1)
				{
					// First check if there aren't any stray equal signs
					//   If there are, the function must be pushed up
					PushUpFunc();
					goto readnext;
				}
				if(pcount>1)
				{
					// There was more than one set of parentheses, so
					//   a function is there to be pushed up. This only
					//   happens if a function is one of the parameters
					//   and there are no parentheses, operations, etc.
					PushUpFunc();
				}
				CallFunction();
				goto readnext;
			}
			if(cl.IsPar()==1)
			{
				// By now all functions present were pushed up
				//   so the next step is pushing up parentheses
				PushUpPars();
				goto readnext;
			}
			if(cl.IsEqual()==1)
			{
				// No more functions or parentheses by now.
				//   Equal sign was found though.
				if(cl.IsAri()==1)
					// Arithmetic symbols also found.
					//   Highest operation to be pushed up.
					PushUpOps();
				else
					// Just a lone equal sign
					//   Process the equality
					MakeEqual();
				goto readnext;
			}
			if(cl.IsAri()==1)
			{
				// Process arithmetic operation
				b=0;
				for(c=1;c<cl.Len();c++)
				{
					// First check if there is more than one arithmetic
					//  symbol there
					if(cl.IsOper(c))
					{
						b++;
						c++;
					}
				}
				if(b==1)
					ProcessAri();
				else
					PushUpOps();
				goto readnext;
			}
			if(a==0)
				// None of the flags are set, so the command
				//   must just be a single word.
				ProcessKeyword();
readnext:
		deb++;
		}
	}
}

void PrepareForParse(ifstream infile, ofstream outfile)
{
	while(infile)
	{
		M_GL;
		if((lcoms==1)&&(cl.txt[0]!='\''))
		{
			// If comment toggle was on, keep original line as comment
			temp=ZSCon("\'")+cl;M_PLT;
		}
		// Place a special copy of the line to report original line
		// if errors occcur during parsing
		temp=ZSCon("\''")+cl;M_PLT;
		M_PLC;
		// Add MEMRESET between every code line
		// It will be used by the parser to clear internal temp
		// memory counter
		PutLine(outfile, "MEMRESET");
	}
}

void PrepareForLink(ifstream infile, ofstream outfile)
{
	// Prepares the file for linking
	while(infile)
	{
		M_GL;
		// MEMRESETS are ignored and thus removed
		if(cl=="MEMRESET");
		else
		{
			if(cl.txt[0]=='\'')
			{
				// Changes Hyperion comments to assambler comments
				// but ignore error-markers, thus removing them
				if(cl.txt[1]=='\'')cl=";";
				else
					cl.txt[0]=';';
			}
			if(cl.txt[0]=='@')
			{
				// Anything follwed by @ is tabbed and left
				cl.txt[0]=' ';
				cl.UnTab();
				cl=ZSCon("     ") + cl;
			}
			if(cl==";");
			else M_PLC;
		}
	}
}

void Linker(ifstream infile, ofstream outfile)
{
	PutLine(outfile,"; Compiled using Hyperion. Hyperion is Copyright 1999 Mikolaj Franaszczuk");
	PutLine(outfile,"");
	PutLine(outfile,"#include \"TI-85.H\"");
	PutLine(outfile,".org 0");
	temp=ZSCon(".db \"") + title + "\", 0";
	// Puts program initializors and title in outfile
	M_PLT;
	PutLine(outfile,"");
	b=-1;
	for(a=1;a<10;a++)
	{
		// Assigns memory addresses to parameter passers
		b=b+2;
		temp=ZSCon("PARV")+ZSCon(a)+ZSCon(" = TEXT_MEM+")+b;M_PLT;
		// Assigns memory addresses to temporary mem spaces
		b=b+2;
		temp=ZSCon("MEM")+ZSCon(a)+ZSCon(" = TEXT_MEM+")+b;M_PLT;
	}
	PutLine(outfile,"");
	b=b+2;
	PutLine(outfile,"");
	// Assigned memory addresses to 1 and 2 byte return variables
	ZString memarea = " = TEXT_MEM+";
	c=0;
	for(a=1;a<=vtotal;a++)
	{
		// Assigns memory addresses to all user variables
		if(var[a].Cmnd("USTR")!=1)
		{
			for(int dd=1;dd<=ilv;dd++)
			{
				// Make sure label variables don't get addresses
				if(lv[dd]==a)goto dontplace;
			}
			if(vstat[a]==1)goto dontplace;
			// Doesn't assign permanent variables here
			b++;
			if(b>100)
			{
				// Reassigns mem space in case of overflow
				b=1;
				if(c==1)
					memarea = " = VIDEO_MEM+";
				else
				{
					c=1;
					memarea = " = TEXT_MEM2+";
				}
			}
			temp=var[a]+memarea+b;M_PLT;
			if(vtype[a]==2)b++;
		}
dontplace:;
	}
	PutLine(outfile,"");
	for(a=1;a<=nstr;a++)
	{
		// Places string correspondences in outfile
		temp = ZSCon("USTR")+ZSCon(a)+ ":";M_PLT;
		temp = ZSCon(".db \"") + ustr[a] + "\", 0";M_PLT;
	}
	PutLine(outfile,"");
	if(chksum==1)
	{
		// Tells Z-Shell not to perform a checksum
		temp = "	 ld HL, $8C40";M_PLT;
		temp = "	 set 0, (HL)";M_PLT;
	}
	PutLine(outfile,"");
	while(infile)
	{
		// Copies the main compiled file
		M_GL;
		if(cl.Cmnd(";FSTART"))
		{
			// A function declaration starts here...
			PutLine(outfile, "");
			temp=cl.MidFT(7,cl.Len());
			a=atoi(temp.txt);
			if(fcalled[a]!=1)
			{
				// Don't copy code for a function that was never 
				//   called within the main program
				while(infile)
				{
					// Keep reading until end of declaration is found
					M_GL;
					if(cl.Cmnd("FEND"))break;
				}
				M_GL;
			}
			PutLine(outfile, "");
		}
		// The following loop ensures all labels are flushed left for TASM
		for(a=1;a<=cl.Len();a++)
			if(cl.txt[a]==':')
				cl.UnTab();
		M_PLC;
	}
	// Places "ret" at end of main program
	PutLine(outfile, "     ret");
	for(a=1;a<=vtotal;a++)
	{
		// Assigns memory addresses to all permament user variables
		if(var[a].Cmnd("USTR")!=1)
		{
			for(int dd=1;dd<=ilv;dd++)
			{
				// Make sure label variables don't get addresses
				if(lv[dd]==a)goto dontplace2;
			}
			if(vstat[a]==0)goto dontplace2;
			// Doesn't assign temporary variables here
			temp = var[a] + ":";M_PLT;
			temp = " .dw 0";M_PLT;
		}
dontplace2:;
	}
	PutLine(outfile, ".end");
	// Added ending .end for TASM
	outfile.close();
	infile.close();
}

void MOverflow()
{
	cout << "\nMemory overflow! Press any key to exit.";
	M_CF;
	getch();
	exit(0);
}

void EndProgram()
{
	M_CF;
	if(preparsing==1)
	{
		cout << "\nThe error occured while parsing the line: ";
		M_CF;
		cl.Disp();
		if(parsing==1)
		{
			cout << "\nThe pre-parsed form of that line was: ";
			M_CF;
			cline.Disp();
		}
	}
	cout << "\n\nCompilation has terminated. Press any key to exit.\n";
	M_CF;
	outfile.close();
	infile.close();
	getch();
	exit(0);
}

void main(int argc,char *argv[])
{
	// Main program structure
	cout << "Hyperion C-to-Z80 Compiler by Mikolaj Franaszczuk. Copyright 1999.\n";
	cout << "Version 1.032. August 31, 1999\n";
	cout << "Visit the Hyperion web page at:\n";
	cout << "http://www.geocities.com/TimesSquare/Arcade/2505/hyperion.html\n";
	cout << "E-mail msf22@cornell.edu with questions or comments.\n";
	M_CF;
	// Read command line arguments and prepare the source file
	SetOps();
	cl.oplist=Ops;
	if(argc<2)
	{
		// Source file name not entered
		cout << "You must enter a source filename.";
		M_CF;
		exit(0);
	}
	debug=false;
	// If the second command line argument is "d" then compilation is
	// put in debug mode (all parsed lines are displayed)
	if(argc==3)
		if(strcmp(argv[2],"d")==0)
			debug=true;
	SFile = argv[1];
	cout << "Press any key to pause compilation at any time.\n";
	cout << "\nCompiling source file: " << argv[1] << '\n' << "Preparing file...\n";M_CF;

	// Main compilation begins here
	PrepareFile(SFile);
	cout << "Processing pre-compiler directives... ";M_CF;

	preparsing=1;
	PreCompile(infile,outfile);
	M_FLIP;

	cout << "\nPreparing file for parsing...\n";M_CF;
	PrepareForParse(infile,outfile);

	nw=1;
	cout << "Parsing ";M_CF;
	parsing=1;
	Parser(infile,outfile);
	M_FLIP;

	cout << "\nPreparing file for linking...\n";M_CF;
	PrepareForLink(infile,outfile);
	M_FLIP;

	outfile.close();
	outfile.open(DFile.txt);
	cout << "Linking...\n";M_CF;
	Linker(infile,outfile);

	cout << "\nCompilation complete. Assembler file is ";M_CF;
	DFile.Disp(2);
	// Deletes the work files
	remove("work2.dat");
	remove("work.dat");
}
